"""Initializer for the hiper board
"""
import logging
import time

from naludaq.controllers.board.hiper import BoardControllerHiper
from naludaq.helpers.decorators import log_function

from .aardvarcv3 import InitAardvarcv3

logger = logging.getLogger("naludaq.init_hiper")


class InitHiper(InitAardvarcv3):
    def __init__(self, board):
        """Initializer for the hiper board.

        Args:
            board (Board): the board to initialize.
        """
        super().__init__(board)

    def run(self) -> bool:
        """Runs the initialization sequence.

        Returns:
            True, always.
        """

        self._reset_digital()

        # Initialize FPGA registers
        self._fpga_init()
        self._i2c_startup()

        # Power on FMC board
        # self._power_toggle(False)
        # self._power_toggle(True)
        time.sleep(0.25)

        # Program the clock
        self._init_clock()

        # Select parallel interface
        magic_coms = [
            "AF040240",
            "AF040040",
            "AF040041",
            "AF040001",
        ]
        self.write_commands(magic_coms)
        time.sleep(0.1)
        # self._set_serial_mode(True)
        time.sleep(0.1)
        # self._set_loopback_enabled(False)
        chips = self.board.params.get("installed_chips", [])
        if len(chips) > 0:
            self._set_chips(chips=chips)
            time.sleep(0.1)
            self._set_txin()
            time.sleep(0.1)

            # Chip-side register startup
            BoardControllerHiper(self.board).analog_startup(chips=chips)
            BoardControllerHiper(self.board).digital_startup(chips=chips)

        if not self._sync_tx_pol():
            logger.error("tx idle detection missmatch")
        if not self._sync_rx_pol():
            logger.error("busy not locked.")

        # External devices
        self._init_dacs()

        return True

    def write_commands(self, commands):
        for cmd in commands:
            self.board.connection.send(cmd)
            time.sleep(0.01)

    def _reset_digital(self):
        """Reset digital side of the ASIC"""
        self.control_write("digrst", True)
        self.control_write("digrst", False)

    @log_function(logger)
    def _reset_sys(self):
        """Reset the ASIC and clear the analog and digital registers"""
        self.control_write("sysrst", True)
        self.control_write("sysrst", False)
        self.control_write("regclr", True)
        self.control_write("regclr", False)

    @log_function(logger)
    def _set_chips(self, chips):
        # chips = [x for x in range(14)]  # self.board.params.get("installed_chips", [])
        chips = sum(1 << chip for chip in chips)
        self.control_write("rxout_en", chips)

    @log_function(logger)
    def _set_txin(self):
        self.control_write("txin_en", False)

    def _sync_tx_pol(self):
        """Set ASIC txin polarity and check if locked

        SET POLARITY TO ASIC TXIN
        IF POLAROITY AND BAUD IS NOT CORRECT IDLE_DETECT WILL BE 0
        IF POLARITY AND BAUD ARE CORRECT IDLE_DETETCT WILL BE 1
        """
        chips = self.board.params.get("installed_chips", [])
        expected_sync = sum(1 << chip for chip in chips)
        for _ in range(5):
            if self.control_registers.read("idle_det")["value"] == expected_sync:
                return True
            time.sleep(0.05)
        return False

    def _sync_rx_pol(self):
        """Set ASIC rxout pol and make sure it's locked

        SET POLARITY TO ASIC RXOUT
        IF POLAROITY AND BAUD IS NOT CORRECT BUSY_LOCKED WILL BE 0
        IF POLARITY AND BAUD ARE CORRECT BUSY_LOCKED WILL BE 1
        """
        chips = self.board.params.get("installed_chips", [])
        expected_sync = sum(1 << chip for chip in chips)
        for _ in range(5):
            if self.control_registers.read("busy_locked")["value"] == expected_sync:
                return True
            time.sleep(0.05)
        return False

    @log_function(logger)
    def _set_loopback_enabled(self, enabled: bool):
        """Enable or disable loopback mode."""
        from naludaq.controllers import get_board_controller

        get_board_controller(self.board).set_loopback_enabled(enabled)



"""Parser for the Oddsock boards

The oddsocks can either have an ASOC or an AODS chip.
The parsing is exactly the same for types of chips.
"""
from logging import getLogger

import numpy as np

from naludaq.helpers.exceptions import BadDataError
from naludaq.parsers.parser import Parser

LOGGER = getLogger("naludaq.oddsock_parser")


class OddsockParser(Parser):
    """Parser for both flavors of the oddsock board."""

    def __init__(self, params) -> None:
        """Create an oddsock parser using the given board params.

        Args:
            params (dict): the board params. If necessary fields are missing,
                reasonable defaults are used instead.
        """
        super().__init__(params)
        self.channels = self.params.get("channels", 8)
        self.channels_per_chip = self.params.get("channels_per_chip", 4)
        self.num_chips = self.params.get("num_chips", 2)
        self.samples = self.params.get("samples", 64)

        self.chip_footer = self.params.get("chip_stop_word", 0xFFFF)
        self.windmask = self.params.get("windmask", 0x0FF)
        self.chanmask = self.params.get("chanmask", 0x300)
        self.chanshift = self.params.get("chanshift", 8)
        self.data_mask = self.params.get("datamask", 0xFFF)

        self.num_window_headers = self.params.get("num_window_headers", 1)
        self.num_chip_headers = self.params.get("chip_headers", 3)

    def _validate_input_package(self, in_data: dict):
        """Run some sanity checks on the raw event before parsing

        Args:
            in_data (dict): raw event with key 'rawdata' (bytes).

        Raises:
            BadDataError: if the event is invalid and cannot be parsed.
        """
        if not isinstance(in_data, dict):
            raise TypeError("Raw event must be a dict")
        raw_data = in_data.get("rawdata", b"")
        if raw_data == b"":
            raise BadDataError("Event contains no data")
        if len(raw_data) % 2 != 0:
            raise BadDataError(
                f"Bad package no {in_data.get('pkg_num', 'unknown')}, "
                f"length is not an even amount of words, "
                f"len {len(in_data['rawdata'])}"
            )

        # data_size strips out chip headers/footers, but not window headers
        data_size = (
            (len(raw_data) // 2)
            - self.num_chip_headers * self.num_chips
            - (self.num_chips - 1)
        )
        if data_size % (self.samples + self.num_window_headers) != 0:
            f"Bad package no {in_data.get('pkg_num', 'unknown')},"
            f"length {len(raw_data)} is not divisible by window size"

    def parse_digital_data_old(self, in_data) -> dict:
        """Parse the raw data from the board.

        Parses the data spit out by the digital readout portion of the firmware

        Args:
            in_data (bytearray): Raw data from the board

        Returns:
            Parsed event as a dict.

        Raises:
            BadDataException if no data is found or if the data contains errors.

        ### Data format
        For each event:
            For each chip:
                For each window:
                    - Chip header (CCCC_XXXX_XXXX_TTTT)
                    - Upper timing header (XXXX_TTTT_TTTTT_TTTT)
                    - Lower timing header (XXXX_TTTT_TTTTT_TTTT)
                    For each channel:
                        - Window header (XXXX_XXCC_WWWW_WWWW)
                        - 64x data packets (XXXX_DDDD_DDDD_DDDD)
                - Chip footer 0xFFFF
            - 0xCAFE
        """
        try:
            self._validate_input_package(in_data)
        except (TypeError, BadDataError):
            raise

        raw_data = np.frombuffer(in_data["rawdata"], dtype=">H")

        # cache a whole bunch of useful values to avoid lookup inside the loops
        channels = self.channels
        channels_per_chip = self.channels_per_chip
        chip_footer = self.chip_footer
        num_window_headers = self.num_window_headers
        num_chip_headers = self.num_chip_headers
        samples = self.samples
        window_stride = num_window_headers + samples
        windmask = self.windmask
        chanshift = self.chanshift
        chanmask = self.chanmask
        datamask = self.data_mask

        curr_event = {
            "window_labels": [[] for _ in range(channels)],
            "data": [[] for _ in range(channels)],
            "start_window": raw_data[num_chip_headers] & windmask,
            "chip_timing": [],
        }

        chip_footer_locs = np.argwhere(raw_data == chip_footer).flatten()
        chip_data_list = np.split(
            raw_data, chip_footer_locs + 1
        )  # +1 includes footer in previous package
        for chip_data in chip_data_list:
            chip_number = chip_data[0] >> 12
            timing = ((chip_data[1] & datamask) << 12) | (chip_data[2] & datamask)
            curr_event["chip_timing"].append(timing)

            window_data = chip_data[
                num_chip_headers:-1
            ]  # -1 to discard chip/event footer
            window_data = np.reshape(
                window_data, (window_data.shape[0] // window_stride, window_stride)
            )
            window_data = np.bitwise_and(window_data, datamask)

            for channel_data in window_data:
                window_number = channel_data[0] & windmask
                channel_number = (
                    (channel_data[0] & chanmask) >> chanshift
                ) + chip_number * channels_per_chip
                curr_event["window_labels"][channel_number].append(window_number)
                curr_event["data"][channel_number].extend(channel_data[1:])

        return curr_event



import binascii
import logging
import select
import socket
import time
from typing import Optional

import numpy as np

from naludaq.board.connections.base_connection import BaseConnection
from naludaq.parsers.answer_parser import get_answer_parser

logger = logging.getLogger(__name__)

_UDP_READ_BUFFER_SIZE = 1500
_UDP_HEADER_BYTES = 16
_WRITE_MESSAGE_NIBBLES = 8
_DEFAULT_TIMEOUT = 0.005


class UDP(BaseConnection):
    def __init__(self, connection_info: dict):
        """Class for communication over an ethernet connection.

        Args:
            connection_info (dict): the connection configuration
        """
        super().__init__()

        # Store connection info
        self._receiver_addr = connection_info["receiver_addr"]
        self._board_addr = connection_info["board_addr"]

        self._model = connection_info["model"]
        self.stop_word = connection_info["stop_word"]
        self._timeout = _DEFAULT_TIMEOUT
        self.read_addr = "AD"
        self.write_addr = "AF"
        self.response_length = 8
        self.parse_answer = get_answer_parser(True)

        self.tx_pause = 0.01
        self.bundle_mode = True

        self._rx_fifo = bytearray()

        self.socket: socket.socket = None

    @property
    def receiver_addr(self) -> tuple:
        """Get/set the user command address.

        Setting the address will close and reopen the connection
        if the connection is already open

        Args:
            addr (tuple): the address
        """
        return self._receiver_addr

    @receiver_addr.setter
    def receiver_addr(self, addr: tuple):
        self._receiver_addr = addr

        # Reopen the socket
        if self.socket:
            self.close()
            self.open()

    @property
    def board_addr(self) -> tuple:
        """Get/set the board command address.

        Args:
            addr (tuple): the address
        """
        return self._board_addr

    @board_addr.setter
    def board_addr(self, addr: tuple):
        self._board_addr = addr
    
    @property
    def type(self):
        return "udp"
    
    @property
    def board_ip(self):
        """Gets the IP address of the board. 
        Alias of UDP.board_addr[0] to use the same naming convention as backend.
        """
        return self.board_addr[0]
    
    @property
    def board_port(self):
        """Gets the IP port of the board. 
        Alias of UDP.board_addr[1] to use the same naming convention as backend.
        """
        return self.board_addr[1]

    @property
    def is_open(self) -> bool:
        """Indicates whether the connection is open."""
        return self.socket is not None

    def set_answer_parser(self, new_firmware: bool = True):
        """Sets the answer parser.

        Args:
            new_firmware (bool): whether to use the parser for new firmware
        """
        self.parse_answer = get_answer_parser(new_firmware)

    def reset_input_buffer(self):
        """Discards any data currently in the input buffer.

        Closes and re-opens the socket.
        """
        self.close()
        self.open()

    def reset_output_buffer(self):
        """Not implemented for ethernet connection."""
        pass

    def __del__(self):
        self.close()

    def open(self):
        """Opens the connection."""
        if not self.socket:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.socket.bind(self._receiver_addr)
            # self.socket.setblocking(False)
            self.socket.settimeout(self._timeout)

    def close(self):
        """Closes the connection."""
        try:
            self.socket.close()
            self.socket = None
        except:
            pass

    def send(
        self, data: str, pause: Optional[float] = None, bundle: Optional[bool] = None
    ):
        """Sends a string of hex characters.

        Args:
            data (str): the data string. Only hex characters are allowed!
            pause (float): the time in seconds to wait after sending
        """
        if bundle is None:
            bundle = self.bundle_mode
        if pause is None:
            pause = self.tx_pause
        # Make sure the length of the input is a multiple of 8
        if len(data) % 8 != 0:
            logger.debug(
                "UDP.send(): need multiples of 8 hex chars, you gave me %s", len(data)
            )
            logger.debug("command:%s", data)
            return

        logger.debug("Sending: %s in %s mode", data, "bundle" if bundle else "single")
        if (len(data) > _WRITE_MESSAGE_NIBBLES) and bundle is False:
            for i in range(0, len(data), _WRITE_MESSAGE_NIBBLES):
                self._send(data[i : i + _WRITE_MESSAGE_NIBBLES], pause)
        else:
            self._send(data, pause)

    def _send(self, data: str, pause: float):
        # Convert hex data string to binary
        # data_binary = binascii.unhexlify(self.sync_word + data)
        data_binary = binascii.unhexlify(data)

        # Send the data over the socket
        self.socket.sendto(data_binary, self._board_addr)
        time.sleep(pause)

    @property
    def in_waiting(self) -> bool:
        """Checks data is waiting in the socket buffer."""
        data_in_socket, _, _ = select.select([self.socket], [], [], 0)
        return len(data_in_socket) != 0

    def read(self, amount: int) -> bytes:
        """Reads bytes, discards first 16 bytes (UDP Header)"""
        # With datagrams, the min amount to read is 16 (header) + 8 (data) = 24
        # But we have to set the buffer to the largest datagram we can receive
        buf_size = max(amount, _UDP_READ_BUFFER_SIZE)
        try:
            data = self.socket.recv(buf_size)
        except Exception as e:
            return b""
        if len(data) < _UDP_HEADER_BYTES:
            return b""
        return data[_UDP_HEADER_BYTES:]

    def read_until(self, stopword: bytes):
        """Pulls data from the connection until a set of characters is found.

        Args:
            stopword (bytes): the sequence of characters that indicate
                when to stop reading.

        Returns:
            A bytearray containing the data, including the trailing stop word.
            The buffer is empty or incomplete if timed out
        """
        buff = bytearray()
        lenstop = len(stopword)
        while buff[-lenstop:] != stopword:
            try:
                next = self.read(24)
                buff.extend(next)
            except Exception as e:
                break
            if len(next) == 0:
                logger.debug(
                    "read_until(): did not end with correct stop word (got %s)",
                    buff[-lenstop:],
                )
                break
        if len(buff) == 0:
            logger.debug("Receive timeout, no data returned")
        return buff

    def read_all(self) -> bytes:
        """Read all waiting bytes"""
        fifo = self._rx_fifo
        while True:
            try:
                data = self.read(_UDP_READ_BUFFER_SIZE)
                fifo.extend(data)
            except Exception as e:
                break
            if len(data) == 0:
                break
        self._rx_fifo = bytearray()
        return fifo

    def receive(self, length: int = -1, timeout: int = 100, raw: bool = False):
        """Receives data from the socket.

        Args:
            length (int): the number of bytes to receive
            timeout (int): the number of milliseconds to wait before giving up
            raw (bool): return bytes (True), or return a hex string (False)

        Returns:
            Hex string or bytes, depending on the `raw` flag, or None if
            no data was received in time.
        """
        return self.read_until(self.stop_word)

    def receive_all(self, timeout: int = 100, raw: bool = False):
        """Receives all waiting data from the socket.

        Args:
            timeout (int): the number of milliseconds to wait before giving up
            raw (bool): return bytes (True), or return a hex string (False)

        Returns:
            Hex string or bytes, depending on the `raw` flag, or None if
            no data was received in time.
        """
        return self.receive(length=self.in_waiting, timeout=timeout, raw=raw)

    def writeReg(self, regNum: int, value: int):
        """Sends a command to write a value to a register

        Args:
            regNum (int): the address of the register to write
            value (int): the value to write
        """
        command = self.write_addr + hex(regNum)[2:].zfill(2) + hex(value)[2:].zfill(4)
        self.send(command)

    def readReg(self, regNum: int, timeout: int = 1000) -> int:
        """Reads a register.

        Args:
            regNum (int): the address of the register to read from
            timeout (int): the number of milliseconds to wait before giving up

        Returns:
            The value held in the register, or None if an error occurred
        """
        cmd_id = self._send_readreg_cmd(regNum)

        # Read a single response from the socket
        while not self.in_waiting and timeout > 0:
            timeout -= 1
            time.sleep(0.001)

        if not self.in_waiting:
            logger.error("UDP.readReg(): Timed out, no data received")
            return None

        buff, _ = self.socket.recvfrom(self.response_length)
        buff = bytearray(buff)

        # Handle the data received
        answer = self.parse_answer(buff)
        if answer["cmd_id"] != cmd_id:
            logger.error(
                "UDP.readReg(): received command ID does not match the one sent"
            )
            return None

        return answer["value"]

    def _send_readreg_cmd(self, regNum) -> int:
        """Sends a command to read a register.

        Args:
            regNum (str, int): the register to read

        Returns:
            The command id used in the sent command
        """
        read_addr = int(self.read_addr, 16)
        if isinstance(regNum, str):
            regNum = int(regNum, 16)

        # Generate and send the command string (4 bytes)
        cmd_id = np.random.randint(0, 2**16)
        command = f"{read_addr:x}{regNum:02x}{cmd_id:04x}"

        self.send(command)

        return cmd_id



"""Factory for the gainstage controller"""


def get_gainstage_controller(board, *args, **kwargs):
    """Return a gainstage controller for the given board"""
    if board.model == "aodsv2_eval":
        from .aodsv2 import GainStageController

        return GainStageController(board)
    elif board.model == "aodsoc_aods":
        from .oddsock_aods import GainStageController

        return GainStageController(board, *args, **kwargs)
    else:
        raise NotImplementedError(
            "No gainstage implemented for %s" % board
        )  # pragma: no cover



"""Base class for all controllers

"""
from abc import ABC


class Controller(ABC):
    def __init__(self, board):
        self._board = None
        self.board = board

    @property
    def board(self):
        return self._board

    @board.setter
    def board(self, board):
        self._board = board



import abc

from naludaq.controllers.controller import Controller


class BaseTIAController(Controller):
    def __init__(self, board):
        """ABC for TIA controllers.

        Args:
            board (Board): the board object.
        """
        super().__init__(board)

    @abc.abstractmethod
    def set_dac_values(self, values: list[int]):
        """Set multiple dac values.

        Args:
            values (list): index in list corresponds to the channel number.
        """



import numpy as np

from naludaq.parsers.headers.base import HeaderParser


class TrbhmHeaderParser(HeaderParser):
    @staticmethod
    def parse_header(event: dict, raw_data: np.ndarray, offset=0):
        """Parse event headers from raw data and store them into an event dict.

        Args:
            event (dict): the event
            raw_data (ndarray): the raw data as words
            offset (int): offset in words of the header
        """
        chip_num = raw_data[offset] & 1

        _store_header_value(event, "channels", chip_num, raw_data[offset + 1])
        _store_header_value(event, "num_winds", chip_num, raw_data[offset + 2])
        _store_header_value(event, "event_id", chip_num, raw_data[offset + 3])
        _store_header_value(event, "prev_final_window", chip_num, raw_data[offset + 4])
        _store_header_value(
            event,
            "trigger_time_ns",
            chip_num,
            (raw_data[offset + 5] << 12) | raw_data[offset + 6],
        )


def _store_header_value(event: dict, key: str, chip_num: int, value: int, num_chips=2):
    headers_dict = event.setdefault("headers", {})
    headers_dict.setdefault(key, [-1] * num_chips)[chip_num] = value



import logging

import numpy as np

from naludaq.helpers.exceptions import BadDataError
from naludaq.parsers.headers import get_header_parser
from naludaq.parsers.parser import Parser

LOGGER = logging.getLogger(__name__)


class TRBHMParser(Parser):
    """Parser for the TRBHM data protocol.

    ```
    Dataformat:
        For EACH chip/window send:

        For UDP send the header in each package
        For UART send one header per event.
        Header
            1x16-bit chip number
            1x16-bit: active channels
            1x16-bit: number of windows in event
            1x16-bit event num
            1x16-bit: last written window (ASIC 1)
            2x16-bit: trigger time_ns (ASIC 2+3)

        Data
                1x channel header: 16-bit
                1x window header 16-bit
                64x data package 16-bit 0000_DDDD_DDDD_DDDD
            … x amount channels
        1x window header (AARDVARCv3 specific digitization timer)

        Footer
        FACE (only for UART, remove FACE in UDP)

        After all data for the event is sent (after final FACE):
        CAFE (Remove in UDP since reassemble happens with event num)

    ```
    Since the event package can contain 1-4 channels of data the window header position
    is hard to determine.

    The solution is to pre-generate all package length combinations and use as a lookup table.
    This allow us to quickly validate packages where channel amount is not ambiguous.

    IN the rare cases it's abiguous there is a test where it checks if window numbers are the same for
    each channel in a window package.

    """

    def __init__(self, params):
        self.params = params

        self._parse_event_headers = get_header_parser(self.params["model"])

        self.test_amount_windows = 3
        self.samples = self.params.get("samples", 64)
        self.windows = self.params.get("windows", 512)
        self.channels = self.params.get("channels", 4)

        self.num_evt_headers = self.params.get("num_evt_headers", 7)
        self.num_channel_headers = self.params.get("num_channel_headers", 2)
        self.num_window_headers = self.params.get("num_window_headers", 2)

        self.num_footer_words = self.params.get("num_footer_words", 1)

        self.chan_step_size = self.samples + self.num_channel_headers

        # Needed?
        self.num_chips = self.params.get("num_chips", 2)

        self.enabled_channels_offset = 1
        self.window_count_offset = 2

        self.trigger_time_offset = 5

        self.connection_type = "uart"

        # Generatethe valid bits dictionary
        # self.valid_bitlengths = self.create_valid_bitlengths()

    def parse(self, raw_data):
        """Parse raw data into an event dictionary

        Args:
            in_data (dict): a dictionary containing at least an entry:
                'rawdata' (bytearray) - the raw data to parse

        Returns:
            An event dictionary

        Raises:
            TypeError if `in_data` is not a dict
            BadDataError if there is no 'rawdata' entry, or if the data does not
                have an even number of bytes
        """

        event = {}
        try:
            # if self.connection_type == 'uart':
            event = self.parse_digital_data_uart(raw_data)
            # else:
            # return self.parse_digital_data_udp(in_data)
        except (TypeError, Exception) as e_msg:
            LOGGER.exception("parse_digital_data failed: %s", e_msg)
            event = raw_data
        else:
            # Add the X-Axis.
            event["time"] = self._add_xaxis_to_event(event)
            event["created_at"] = raw_data.get("created_at", None)
            event["pkg_num"] = raw_data.get("pkg_num", 0)
            event["event_num"] = raw_data.get("pkg_num", 0)
        event["name"] = None

        return event

    def parse_digital_data_uart(self, in_data):
        """Parse raw data into an event dictionary

        Args:
            in_data (dict): a dictionary containing at least an entry:
                'rawdata' (bytearray) - the raw data to parse

        Returns:
            An event dictionary

        Raises:
            TypeError if `in_data` is not a dict
            BadDataError if there is no 'rawdata' entry, or if the data does not
                have an even number of bytes
        """
        try:
            # make sure package contains data and there's an even amount of bytes to combine.
            self._validate_input_package(in_data)
        except:
            raise

        # Parsing variables
        num_chips = self.num_chips

        num_evt_headers = self.num_evt_headers
        num_window_headers = 2  # Move to params
        samples = self.samples
        channels_per_chip = 4  # Move to params

        channel_step_size = self.chan_step_size
        num_footer_words = self.num_footer_words

        channel_footers = 1  # AARDVARCv3 window timing footer
        # Split the event into the data for each chip
        data = np.frombuffer(in_data["rawdata"], dtype=">H")

        # enabled_channels = data[enabled_channels_offset] & enabled_channels_mask
        num_channels = 4  # bin(enabled_channels).count('1')  # replace with dict
        # Set up the event dict
        event = {
            "window_labels": [[] for _ in range(num_channels * num_chips)],
            "data": [[] for _ in range(num_channels * num_chips)],
            "start_window": data[num_evt_headers],
        }

        out_timing_data = []
        step_size = channel_step_size * num_channels + self.num_window_headers

        window_size = num_channels * channel_step_size + num_footer_words
        # num_windows = (len(data) - num_evt_headers) // window_size

        # Only add timing information for the first chip (all chips is too many windows)
        event["window_timings"] = data[window_size - 1]

        step_size = (samples + num_window_headers) * num_channels + channel_footers
        header_steps = range(0, len(data) - step_size, step_size + num_evt_headers + 1)
        self._parse_event_headers(event, data)
        self._parse_event_headers(event, data, step_size + num_evt_headers + 1)

        for idx in header_steps:
            window_steps = range(
                idx + num_evt_headers, idx + step_size - 1, samples + num_window_headers
            )
            chipnum = data[idx]

            for inner_idx in window_steps:
                # First two words are window and channel
                window = data[inner_idx]
                channel = data[inner_idx + 1] + (chipnum * channels_per_chip)
                event["window_labels"][channel].append(window)
                # Next 64 words are data.
                event["data"][channel].extend(
                    data[inner_idx + 2 : inner_idx + channel_step_size]
                )
            # Last word is the window timing
            out_timing_data.append(data[idx + step_size - 1])

        return event

    def _validate_input_package(self, in_data):
        """Validate the input data.

        Raises:
            BadDataException if:
            - there is no data.
            - there is not a en event amount of bytes, since 8-bits are combined to 16-bit words.
            - the data is not divisable by sample_num + window_headers.
        """
        # Correct format
        if not isinstance(in_data, dict):
            raise TypeError("raw_data is not a dict it's a %s", type(in_data))

        # No data
        if in_data.get("rawdata", None) == b"":
            raise BadDataError("No Data Found")

        # Edge case, data doesn't contain valid words.
        if len(in_data["rawdata"]) % 2 != 0:
            raise BadDataError(
                f"Bad package no {in_data.get('pkg_num', 'unknown')}, "
                f"length is not an even amount of words, "
                f"len {len(in_data.get('rawdata', {}))}"
            )

    def _add_xaxis_to_event(self, event):
        """Adds an x-axis to the data.

        Based on the amount of channels and samples it will add a timeaxis
        per channel to the event dict.

        Returns:
            numpy array with sample numbers for each channel.
        """
        times = list()
        samples = self.samples
        channels = len(event["data"])
        start_window = event["start_window"]
        for chan in range(0, channels):
            # LOGGER.debug("winds=%s", event["window_labels"][chan])

            winds = np.array(event["window_labels"][chan], dtype=int)
            winds += self.windows * (winds < start_window)
            winds -= start_window

            tim = (samples * np.repeat(winds, samples)) + (
                np.ones((len(winds), samples), dtype=int)
                * np.arange(0, samples, dtype=int)
            ).flatten()

            times.append(tim)

        return times



# Set default logging handler to avoid "No handler found" warnings.
import logging
from logging import NullHandler

from naludaq._version import __version__

from .naludaq import HiperDaq, NaluDaq, UpacDaq

logging.getLogger(__name__).addHandler(NullHandler())



from naludaq.devices.spi_daisy_chain import SPIDaisyChain
from enum import Enum, auto


class SPIBUS_t(Enum):
    """SPI bus type enum."""

    DAISY_CHAIN = auto()
    SPI = auto()


class SPIConnection:
    """SPI connection class.

    Used as a baseclass for board specific SPI implementations.

    Hadnles multiple buses and devices.

    Args:
        board (Board): The board object.

    Attributes:
        board (Board): The board object.
        spi_bus (DaisyChain): The SPI bus object.
    """

    def __init__(self, board):
        self.board = board
        # self.spi_bus = board.params.get("spi", {}).get("bus", 0)
        self.spi_bus = {}

    @property
    def initialized(self):
        """Check if the SPI connection is initialized."""
        return all([bus.initialized for bus in self.spi_bus.values()])

    def init(self):
        """Initialize the SPI connection."""
        for bus in self.spi_bus.values():
            bus.init()

    def add_bus(self, bus_id: int, bus_type: SPIBUS_t):
        """Add a bus to the SPI connection.

        A spi bus is either a daisy chain or a normal SPI bus.

        Args:
            bus_id (int): The bus id.
            bus_config (int): The bus configuration.
        """
        if bus_type == SPIBUS_t.DAISY_CHAIN:
            self.spi_bus[bus_id] = SPIDaisyChain(self.board, bus_id)
        else:
            raise NotImplementedError("Only daisy chain is currently supported.")

    def add_device(self, device, bus_id: int, device_id: int):
        """Add a device to the SPI bus.

        Args:
            device (Device): The device object.
            bus_id (int): The bus id to add the device to.
        """
        if bus_id not in self.spi_bus:
            raise AttributeError(f"Bus {bus_id} not in SPI connection")

        self.spi_bus[bus_id][device_id] = device

    def write(self, message, device_id, bus_id: "int | None" = None):
        """Write a message to a device on the SPI bus.

        Args:
            message: The message to write.
            device_id: The device id on the SPI bus.
            bus_id: The bus id to write to. If None, the default bus will be used.
        """
        self.spi_bus[bus_id].add_message(message, device_id)
        self.spi_bus[bus_id].push_messages()

    def _generate_write(self, message, device_id, bus_id):
        """Generate the message for the SPI bus.

        Args:
            message: The message to write.
            device_id: The device id on the SPI bus.
            bus_id: The bus id to write to. If None, the default bus will be used.

        Returns:
            str: The generated message.
        """
        msg = self.spi_bus[bus_id]._generate_message(message, device_id, bus_id)
        return msg

    def __getitem__(self, key: int) -> SPIDaisyChain:
        return self.spi_bus[key]

    def __setitem__(self, key: int, value: SPIDaisyChain):
        self.spi_bus[key] = value

    def __len__(self):
        return len(self.spi_bus)

    def __iter__(self):
        for each in self.spi_bus.values():
            yield each

    # return iter(self.spi_bus)
